@@ -4,6 +4,7 @@ import pytest
 import yaml
 
 from ..config import (
+    _resolve_dsn,
     ConfigError,
     DB_ERRORS_METRIC_NAME,
     GLOBAL_METRICS,
@@ -37,8 +38,8 @@ def config_full():
 
 
 @pytest.fixture
-def write_config(tmpdir):
-    path = tmpdir / "config"
+def write_config(tmp_path):
+    path = tmp_path / "config"
 
     def write(data):
         path.write_text(yaml.dump(data), "utf-8")
@@ -179,6 +180,43 @@ class TestLoadConfig:
             load_config(fd, logger)
         assert str(err.value) == 'Invalid database DSN: "invalid"'
 
+    def test_load_databases_dsn_details(self, logger, write_config):
+        """The DSN can be specified as a dictionary."""
+        config = {
+            "databases": {
+                "db1": {
+                    "dsn": {
+                        "dialect": "sqlite",
+                        "database": "/path/to/file",
+                    }
+                }
+            },
+            "metrics": {},
+            "queries": {},
+        }
+        config_file = write_config(config)
+        with config_file.open() as fd:
+            config = load_config(fd, logger)
+        assert config.databases["db1"].dsn == "sqlite:///path/to/file"
+
+    def test_load_databases_dsn_details_only_dialect(self, logger, write_config):
+        """Only the "dialect" key is required in DSN."""
+        config = {
+            "databases": {
+                "db1": {
+                    "dsn": {
+                        "dialect": "sqlite",
+                    }
+                }
+            },
+            "metrics": {},
+            "queries": {},
+        }
+        config_file = write_config(config)
+        with config_file.open() as fd:
+            config = load_config(fd, logger)
+        assert config.databases["db1"].dsn == "sqlite://"
+
     def test_load_databases_dsn_invalid_env(self, logger, write_config):
         """An error is raised if the DSN from environment is invalid."""
         config = {
@@ -203,6 +241,36 @@ class TestLoadConfig:
             load_config(fd, logger, env={})
         assert str(err.value) == 'Undefined variable: "FOO"'
 
+    def test_load_databases_dsn_from_file(self, tmp_path, logger, write_config):
+        """The database DSN can be loaded from a file."""
+        dsn = "sqlite:///foo"
+        dsn_path = tmp_path / "dsn"
+        dsn_path.write_text(dsn)
+        config = {
+            "databases": {"db1": {"dsn": f"file:{dsn_path}"}},
+            "metrics": {},
+            "queries": {},
+        }
+        config_file = write_config(config)
+        with config_file.open() as fd:
+            config = load_config(fd, logger)
+        assert config.databases["db1"].dsn == dsn
+
+    def test_load_databases_dsn_from_file_not_found(self, logger, write_config):
+        """An error is raised if the DSN file can't be read."""
+        config = {
+            "databases": {"db1": {"dsn": "file:/not/found"}},
+            "metrics": {},
+            "queries": {},
+        }
+        config_file = write_config(config)
+        with pytest.raises(ConfigError) as err, config_file.open() as fd:
+            load_config(fd, logger)
+        assert (
+            str(err.value)
+            == 'Unable to read dsn file : "/not/found": No such file or directory'
+        )
+
     def test_load_databases_labels(self, logger, write_config):
         """Labels can be defined for databases."""
         config = {
@@ -336,6 +404,7 @@ class TestLoadConfig:
 
     @pytest.mark.parametrize("global_name", list(GLOBAL_METRICS))
     def test_load_metrics_reserved_name(self, config_full, write_config, global_name):
+        """An error is raised if a reserved label name is used."""
         config_full["metrics"][global_name] = {"type": "counter"}
         config_file = write_config(config_full)
         with pytest.raises(ConfigError) as err, config_file.open() as fd:
@@ -543,6 +612,7 @@ class TestLoadConfig:
     def test_load_queries_section_invalid_timeout(
         self, logger, config_full, write_config, timeout, error_message
     ):
+        """An error is raised if query timeout is invalid."""
         config_full["queries"]["q"]["timeout"] = timeout
         config_file = write_config(config_full)
         with pytest.raises(ConfigError) as err, config_file.open() as fd:
@@ -589,6 +659,7 @@ class TestLoadConfig:
     def test_configuration_warning_unused(
         self, caplog, logger, config_full, write_config
     ):
+        """A warning is logged if unused entries are present in config."""
         config_full["databases"]["db2"] = {"dsn": "sqlite://"}
         config_full["databases"]["db3"] = {"dsn": "sqlite://"}
         config_full["metrics"]["m2"] = {"type": "gauge"}
@@ -676,6 +747,7 @@ class TestLoadConfig:
         )
 
     def test_load_queries_no_metrics(self, logger, config_full, write_config):
+        """An error is raised if no metrics are configured."""
         config_full["queries"]["q"]["metrics"] = []
         config_file = write_config(config_full)
         with pytest.raises(ConfigError) as err, config_file.open() as fd:
@@ -683,6 +755,7 @@ class TestLoadConfig:
         assert str(err.value) == "Invalid config at queries/q/metrics: [] is too short"
 
     def test_load_queries_no_databases(self, logger, config_full, write_config):
+        """An error is raised if no databases are configured."""
         config_full["queries"]["q"]["databases"] = []
         config_file = write_config(config_full)
         with pytest.raises(ConfigError) as err, config_file.open() as fd:
@@ -690,3 +763,55 @@ class TestLoadConfig:
         assert (
             str(err.value) == "Invalid config at queries/q/databases: [] is too short"
         )
+
+
+class TestResolveDSN:
+    def test_all_details(self):
+        """The DSN can be specified as a dictionary."""
+        details = {
+            "dialect": "postgresql",
+            "user": "user",
+            "password": "secret",
+            "host": "dbsever",
+            "port": 1234,
+            "database": "mydb",
+            "options": {"foo": "bar", "baz": "bza"},
+        }
+        assert (
+            _resolve_dsn(details, {})
+            == "postgresql://user:secret@dbsever:1234/mydb?foo=bar&baz=bza"
+        )
+
+    def test_db_as_path(self):
+        """If the db name is a path, it's treated accordingly."""
+        details = {
+            "dialect": "sqlite",
+            "database": "/path/to/file",
+        }
+        assert _resolve_dsn(details, {}) == "sqlite:///path/to/file"
+
+    def test_encode_user_password(self):
+        """The user and password are encoded."""
+        details = {
+            "dialect": "postgresql",
+            "user": "us%r",
+            "password": "my pass",
+            "host": "dbsever",
+            "database": "/mydb",
+        }
+        assert _resolve_dsn(details, {}) == "postgresql://us%25r:my+pass@dbsever/mydb"
+
+    def test_encode_options(self):
+        """Option parmaeters are encoded."""
+        details = {
+            "dialect": "postgresql",
+            "database": "/mydb",
+            "options": {
+                "foo": "a value",
+                "bar": "another/value",
+            },
+        }
+        assert (
+            _resolve_dsn(details, {})
+            == "postgresql:///mydb?foo=a+value&bar=another%2Fvalue"
+        )
